/*
	Modified from 
	http://www.graphics.cornell.edu/~bjw/
*/

#include "define.h"

#define MINRUNLENGTH 4

void RGBE_WriteHeader(FILE *fp, int width, int height);
void RGBE_WritePixels_RLE(FILE *fp, irradiance *data, int scanline_width, int num_scanlines);
void RGBE_WriteBytes_RLE(FILE *fp, unsigned char *buffer, int numbytes);
inline void float2rgbe(unsigned char rgbe[4], double red, double green, double blue)
{
	double v;
	int e;

	v = red;
	if (green > v) v = green;
	if (blue > v) v = blue;
	if (v < 1e-32) 
	{
		rgbe[0] = rgbe[1] = rgbe[2] = rgbe[3] = 0;
	}
	else 
	{
		v = frexp(v,&e) * 256.0/v;
		rgbe[0] = (unsigned char) (red * v);
		rgbe[1] = (unsigned char) (green * v);
		rgbe[2] = (unsigned char) (blue * v);
		rgbe[3] = (unsigned char) (e + 128);
	}
}

void hdr_write(char *filename, hdr *h)
{
	FILE *out = fopen(filename, "wb");
	RGBE_WriteHeader(out, h->xresolution, h->yresolution);
	RGBE_WritePixels_RLE(out, h->s, h->xresolution, h->yresolution);
	fclose(out);
}

void RGBE_WritePixels_RLE(FILE *fp, irradiance *data, int scanline_width, int num_scanlines)
{
	unsigned char rgbe[4];
	unsigned char *buffer;
	int i, idx;

	if ((scanline_width < 8)||(scanline_width > 0x7FFF))
	{		
		printf("Image size is too big or too small\n");
		exit(1);
	}
	buffer = (unsigned char *)malloc(sizeof(unsigned char)*4*scanline_width);
	if (buffer == NULL) 
	{
		printf("Memory allocation fails\n");
		exit(1);
	}

	idx = 0;
	while(num_scanlines-- > 0)
	{
		rgbe[0] = 2;
		rgbe[1] = 2;
		rgbe[2] = scanline_width >> 8;
		rgbe[3] = scanline_width & 0xFF;
		fwrite(rgbe, sizeof(rgbe), 1, fp);

		idx = (num_scanlines)*scanline_width;

		for(i = 0; i < scanline_width; i++) 
		{
			float2rgbe(rgbe, data[idx].r, data[idx].g, data[idx].b);
			buffer[i]					= rgbe[0];
			buffer[i+scanline_width]	= rgbe[1];
			buffer[i+2*scanline_width]	= rgbe[2];
			buffer[i+3*scanline_width]	= rgbe[3];
			idx++;
		}

		/* write out each of the four channels separately run length encoded */
		/* first red, then green, then blue, then exponent */
		for(i = 0; i < 4; i++) 
		{
			RGBE_WriteBytes_RLE(fp, &buffer[i*scanline_width], scanline_width);
		}
	}
	free(buffer);
}

void RGBE_WriteHeader(FILE *fp, int width, int height)
{
	char *programtype = "RADIANCE";

	/* The #? is to identify file type, the programtype is optional. */
	fprintf(fp,"#?%s\n",programtype);	
	fprintf(fp,"FORMAT=32-bit_rle_rgbe\n\n");	
	fprintf(fp, "+Y %d +X %d\n", height, width);
}

void RGBE_WriteBytes_RLE(FILE *fp, unsigned char *data, int numbytes)
{
	int cur, beg_run, run_count, old_run_count, nonrun_count;
	unsigned char buf[2];

	cur = 0;
	while(cur < numbytes) 
	{
		beg_run = cur;
		/* find next run of length at least 4 if one exists */
		run_count = old_run_count = 0;
		while((run_count < MINRUNLENGTH) && (beg_run < numbytes)) 
		{
			beg_run += run_count;
			old_run_count = run_count;
			run_count = 1;
			while((data[beg_run] == data[beg_run + run_count]) 
				&& (beg_run + run_count < numbytes) && (run_count < 127))
				run_count++;
		}
		/* if data before next big run is a short run then write it as such */
		if ((old_run_count > 1)&&(old_run_count == beg_run - cur)) 
		{
			buf[0] = 128 + old_run_count;   /*write short run*/
			buf[1] = data[cur];
			fwrite(buf,sizeof(buf[0])*2,1,fp);
			cur = beg_run;
		}
		/* write out bytes until we reach the start of the next run */
		while(cur < beg_run) 
		{
			nonrun_count = beg_run - cur;
			if (nonrun_count > 128) 
				nonrun_count = 128;
			buf[0] = nonrun_count;
			fwrite(buf,sizeof(buf[0]),1,fp);
			fwrite(&data[cur],sizeof(data[0])*nonrun_count,1,fp);
			cur += nonrun_count;
		}
		/* write out next run if one was found */
		if (run_count >= MINRUNLENGTH) 
		{
			buf[0] = 128 + run_count;
			buf[1] = data[beg_run];
			fwrite(buf,sizeof(buf[0])*2,1,fp);
			cur += run_count;
		}
	}
}